//--------------------------------------------------------------------------------------------------------------------//
//                                                QUATERNION CLASS                                                    //
// 26/08/00, Mythos                                                                                                   //
// 03/10/00: Alexis'modifs (TR interpolation)                                                                         //
//--------------------------------------------------------------------------------------------------------------------//

#ifndef   _QUATERNION_H_
#define   _QUATERNION_H_

//----------------------------------------------- INCLUDES -----------------------------------------------------------//

#include  "VectMat.h"                                       // vectors & matrices

//----------------------------------------------- CLASSES ------------------------------------------------------------//

//--------------------------------------------------------------------------------------------------------------------//
//                                                QUATERNION                                                          //
//--------------------------------------------------------------------------------------------------------------------//

class CQuaternion
  {
// public methods

  public:

    // constructors

    inline          CQuaternion         (void);
    inline          CQuaternion         (const CVect3D&     v3Vector,const float fScalar);
    inline          CQuaternion         (const CVect4D&     v4Quat);                      // x/y/z=vector, w=scalar
    inline          CQuaternion         (const float        fX,const float fY,const float fZ,const float fW);
    inline          CQuaternion         (const CQuaternion& quatQ);

    // assignment operators

    CQuaternion&       operator       = (const CQuaternion& quatQ);                       // assignment
    CQuaternion&       operator      += (const CQuaternion& quatQ);                       // incrementation
    CQuaternion&       operator      -= (const CQuaternion& quatQ);                       // decrementation
    CQuaternion&       operator      *= (const CQuaternion& quatQ);                       // multiplication
    CQuaternion&       operator      *= (const float        fK);                          // multiplication by a constant
    CQuaternion&       operator      /= (const float        fK);                          // division       by a constant

    // friend operators

    friend CQuaternion inline operator -(const CQuaternion& quatQ);                                 // -Q
    friend CQuaternion inline operator *(const float        fK,    const CQuaternion& quatQ);       // k*Q 
    friend CQuaternion inline operator *(const CQuaternion& quatQ, const float        fK);          // Q*k
    friend CQuaternion inline operator /(const CQuaternion& quatQ, const float        fK);          // Q/k
    friend CQuaternion inline operator +(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1+Q2
    friend CQuaternion inline operator -(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1-Q2
    friend CQuaternion inline operator *(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1*Q2
    friend bool        inline operator==(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1==Q2
    friend bool        inline operator!=(const CQuaternion& quatQ1,const CQuaternion& quatQ2);      // Q1!=Q2

    friend ostream&        operator  << (ostream& s,        CQuaternion& quatQ);          // output
    friend ostream&        operator  << (ostream& s,const   CQuaternion& quatQ);          // output

    // special functions

    friend CQuaternion     quatSlerp    (const CQuaternion& quatQ1,const CQuaternion& quatQ2,const float fT);
           CQuaternion     quatConjugate(void) const;
    float  inline          fLength      (void) const;
    float  inline          fLength2     (void) const;

    // conversions

    void            vFromAxisAngle      (const CVect3D&      v3Axis,const float  fAngle);
    void            vToAxisAngle        (      CVect3D*     pv3Axis,      float* pfAngle);

    void            vFromRotation       (const CMat3x3&      m3M);
    void            vFromRotation       (const CMat4x4&      m4M);
    void            vToRotation         (      CMat3x3*     pm3M);
    void            vToRotation         (      CMat4x4*     pm4M);

    // get/set

    inline float    fGetScalar          (void);
    inline CVect3D& v3GetVector         (void);

    inline void     vSetScalar          (const float        fScalar);
    inline void     vSetVector          (const CVect3D&     v3Vector);
    inline void     vSetVector          (const float        fX,const float fY,const float fZ);
    inline void     vSet                (const float        fX,const float fY,const float fZ,const float fW);
    inline void     vSet                (const CVect4D&     v4Quat);
    inline void     vSet                (const CQuaternion& quatQ);

    // friend functions

    friend void     vAdd                (CQuaternion* pquatQ, const CQuaternion& quatQ1);                               // Q+=Q1
    friend void     vSub                (CQuaternion* pquatQ, const CQuaternion& quatQ1);                               // Q-=Q1
    friend void     vMul                (CQuaternion* pquatQ, const CQuaternion& quatQ1);                               // Q*=Q1
    friend void     vMul                (CQuaternion* pquatQ, const float        fK);                                   // Q*=k
    friend void     vDiv                (CQuaternion* pquatQ, const float        fK);                                   // Q/=k

    friend void     vNeg                (CQuaternion* pquatQ, const CQuaternion& quatQ1);                               // Q=-Q1
    friend void     vMul                (CQuaternion* pquatQ, const CQuaternion& quatQ1,const float        fK);         // Q=Q1*k
    friend void     vDiv                (CQuaternion* pquatQ, const CQuaternion& quatQ1,const float        fK);         // Q=Q1/k
    friend void     vAdd                (CQuaternion* pquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2);     // Q=Q1+Q2
    friend void     vSub                (CQuaternion* pquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2);     // Q=Q1-Q2

    friend void     vMul                (CQuaternion* pquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2);     // Q=Q1*Q2

// protected members

  protected:
                                                                                          // ! MUST be the same size as a CVect4D
    CVect3D         m_v3Vector;                                                           // imaginary part
    float           m_fScalar;                                                            // real      part
  };

//----------------------------------------------- INLINES ------------------------------------------------------------//

#include  "Quaternion.inl"

//--------------------------------------------------------------------------------------------------------------------//

#endif // _QUATERNION_H_
